#include "../../plugin.hpp"
#include <imagehlp.h>
#include "avl_sid.hpp"

typedef BOOL (WINAPI *LookupAccountSidCallback)(LPCSTR,PSID,LPSTR,PDWORD,LPSTR,PDWORD,PSID_NAME_USE);
LookupAccountSidCallback OldLookupAccountSidA=NULL;
DWORD *PatchAddressA=NULL;
avl_sid_tree *sids=NULL;

static BOOL WINAPI NewLookupAccountSidA(LPCSTR lpSystemName,PSID Sid,LPSTR Name,PDWORD cbName,LPSTR DomainName,PDWORD cbDomainName,PSID_NAME_USE peUse)
{
  if(OldLookupAccountSidA)
  {
    BOOL result=FALSE;
    avl_sid_data *cache_entry=sids->query(Sid);
    if(cache_entry)
    {
      if(cache_entry->AccounName.length()<*cbName)
      {
        strcpy(Name,(char *)cache_entry->AccounName.get());
        if(cache_entry->DomainName.length()<*cbDomainName)
        {
          strcpy(DomainName,(char *)cache_entry->DomainName.get());
          result=TRUE;
        }
        else *cbDomainName=cache_entry->DomainName.length()+1;
      }
      else *cbName=cache_entry->AccounName.length()+1;
      if(!result) SetLastError(ERROR_INSUFFICIENT_BUFFER);
      return result;
    }
    else
    {
      result=OldLookupAccountSidA(lpSystemName,Sid,Name,cbName,DomainName,cbDomainName,peUse);
      DWORD Error=GetLastError();
      if(result)
      {
        avl_sid_data *Add=new avl_sid_data(Sid,(const unsigned char *)Name,(const unsigned char *)DomainName);
        sids->insert(Add);
      }
      else if(Error==ERROR_NONE_MAPPED)
      {
        avl_sid_data *Add=new avl_sid_data(Sid);
        sids->insert(Add);
      }
      SetLastError(Error);
      return result;
    }
  }
  else return FALSE;
}

static bool UpdateDWORD(DWORD *ptr,DWORD data)
{
  bool res=false;
  MEMORY_BASIC_INFORMATION meminfo;
  if(VirtualQuery(ptr,&meminfo,sizeof(meminfo)))
  {
    DWORD oldaccess;
    if(VirtualProtect(meminfo.BaseAddress,meminfo.RegionSize,PAGE_READWRITE,&oldaccess))
    {
      *ptr=data;
      VirtualProtect(meminfo.BaseAddress,meminfo.RegionSize,oldaccess,&oldaccess);
      res=true;
    }
  }
  return res;
}

static void PatchAPI(const char *APIName,void *NewAPI,LookupAccountSidCallback *OldAPI,DWORD **PatchAddress)
{
  HMODULE h=GetModuleHandle(NULL);
  if(h)
  {
    ULONG import_size;
    PIMAGE_IMPORT_DESCRIPTOR data=(PIMAGE_IMPORT_DESCRIPTOR)ImageDirectoryEntryToData(h,TRUE,IMAGE_DIRECTORY_ENTRY_IMPORT,&import_size);
    if(data)
    {
      while(data->Name)
      {
        char *ModName=(char *)h+data->Name;
        if(!_stricmp(ModName,"advapi32.dll"))
        {
          PIMAGE_THUNK_DATA thunk1=(PIMAGE_THUNK_DATA)((char *)h+(unsigned long)data->OriginalFirstThunk);
          PIMAGE_THUNK_DATA thunk2=(PIMAGE_THUNK_DATA)((char *)h+(unsigned long)data->FirstThunk);
          if(thunk1&&thunk2)
          {
            while(thunk1->u1.AddressOfData&&thunk2->u1.Function)
            {
              PIMAGE_IMPORT_BY_NAME FuncData=(PIMAGE_IMPORT_BY_NAME)((char *)h+(unsigned long)thunk1->u1.AddressOfData);
              if(!_stricmp((char *)FuncData->Name,APIName))
              {
                *OldAPI=(LookupAccountSidCallback)thunk2->u1.Function;
                if(!UpdateDWORD((DWORD *)&(thunk2->u1.Function),(DWORD)NewAPI)) *OldAPI=NULL;
                else *PatchAddress=(DWORD *)&(thunk2->u1.Function);
                break;
              }
              thunk1++; thunk2++;
            }
          }
          break;
        }
        data++;
      }
    }
  }
}

void WINAPI SetStartupInfo(const struct PluginStartupInfo *Info)
{
  (void)Info;
  PatchAPI("LookupAccountSidA",(void *)&NewLookupAccountSidA,&OldLookupAccountSidA,&PatchAddressA);
  sids=new avl_sid_tree();
}

void WINAPI ExitFAR()
{
  if(OldLookupAccountSidA)
  {
    if(UpdateDWORD(PatchAddressA,(DWORD)OldLookupAccountSidA))
    {
      PatchAddressA=NULL;
      OldLookupAccountSidA=NULL;
    }
  }
  delete sids;
}

void WINAPI GetPluginInfo(struct PluginInfo *Info)
{
  Info->StructSize=sizeof(*Info);
  Info->Flags=PF_PRELOAD;
}

#ifdef __cplusplus
extern "C"{
#endif
  bool WINAPI DllMainCRTStartup(HANDLE hDll,DWORD dwReason,LPVOID lpReserved);
#ifdef __cplusplus
};
#endif

bool WINAPI DllMainCRTStartup(HANDLE hDll,DWORD dwReason,LPVOID lpReserved)
{
  (void)hDll;
  (void)dwReason;
  (void)lpReserved;
  return true;
}
